"""
Models module for presets.

This module contains model definitions and references.
"""

from .cirad_tf import *
# from .cirad_torch import *
from .generic_tf import *
# from .generic_torch import *

__all__ = []
