"""
Splitters module for presets.

This module contains data splitting presets.
Note: This directory is currently empty.
"""
from .splitters import KennardStoneSplitter, SPXYSplitter, KMeansSplitter, SPlitSplitter, SystematicCircularSplitter, KBinsStratifiedSplitter
__all__ = ["KennardStoneSplitter", "SPXYSplitter", "KMeansSplitter", "SPlitSplitter", "SystematicCircularSplitter", "KBinsStratifiedSplitter"]
